/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import java.util.*;
import java.io.*;

public class CurvedOffline extends JPanel implements Runnable
{
	//private attributes

	//background color
	private Color bgColor=Color.black;
	//determines whether game in the running state or not
	private boolean running;
	//2-dimensional array holding the positions of he playfield which are occupied
	private boolean playField[][];
	//array with the snakes
	private Snake[] snakes;
	//key events
	private boolean left[];
	private boolean right[];
	//the scores of the players
	private int scores[];
	//the colors of the snakes
	private Color[] colors;
	//key configuration
	private int[] leftKeys;
	private int[] rightKeys;
	private int readyKey=KeyEvent.VK_SPACE;
	private int quitKey=KeyEvent.VK_ESCAPE;
	//counter to wait for some loops at the start of each round
	private int waitCounter;
	//fullscreen or not?
	private boolean fullscreen;
	//# of players
	private int players;
	//# of players alive
	private int alive;
	//score needed to win the game
	private int maxScore;
	//the winner of the game
	private int winner;
	//gamespeed (0=low,1=normal,2=high)
	private int speed;
	//turnspeed (0=low,1=normal,2=high)
	private int turnSpeed;
	//amount of holes (0=low,1=normal,2=high)
	private int holes;

	//--game status flags--
	//determines whether the playfield needs to be redrawn or not
	private boolean redraw;
	//game started flag
	private boolean started;
	//game over flag
	private boolean theEnd;

	//old (default) displaymode
	private DisplayMode old;
	//game properties
	private Properties props;

	private JFrame window;

	//constructor
	public CurvedOffline()
	{
		//load properties
		loadProperties();

		setPreferredSize(new Dimension(800,600));
		setBackground(bgColor);

		//show the game JPanel in fullscreen mode
		if(fullscreen)
		{
			//create the window
			window=new JFrame("~ CurveD ~");
			//hide frame and titlebar
			window.setUndecorated(true);
			//add the game JPanel
			window.getContentPane().add(this);
			//size
			window.pack();
			//fullscreen mode
			GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
        	GraphicsDevice device = env.getDefaultScreenDevice();
        	device.setFullScreenWindow(window);
        	old=device.getDisplayMode();
        	device.setDisplayMode(new DisplayMode(800, 600, old.getBitDepth(), old.getRefreshRate()));
		}
		else
		//show the game JPanel in windowed mode
		{
			//create the window
			window=new JFrame("~ CurveD ~");
			//add the game JPanel
			window.getContentPane().add(this);
			//size
			window.pack();
			window.setResizable(false);
			window.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			//center
			window.setLocationRelativeTo(null);
			//show the window
			window.show();
		}

		//hide mouse cursor
		BufferedImage cur = new BufferedImage(1, 1, BufferedImage.TYPE_4BYTE_ABGR);
		setCursor(getToolkit().createCustomCursor(cur,new Point(0,0),""));

		//register keylistener to handle keyevents
		this.addKeyListener(
			new KeyAdapter()
			{
				//key pressed
				public void keyPressed (KeyEvent e)
				{
					//--movement keys--

					//movement keys may only be registered when the game
					//is in the running state
					if(running)
					{
						//movement key pressed --> set pressed flags
						for(int i=0;i<players;i++)
						{
							if (e.getKeyCode() == leftKeys[i])
							{
								left[i]=true;
							}
							else if (e.getKeyCode() == rightKeys[i])
							{
								right[i]=true;
							}
						}
					}
					//--other keys--

					//ready key pressed
					else if (e.getKeyCode() == readyKey)
					{
						if(!started || theEnd)
						{
							//start a new game
							started=false;
							restart();
						}
						else
						{
							//start a new round
							nextRound();
						}
					}
					//quit key pressed
					if(e.getKeyCode()== quitKey)
					{
						if(fullscreen)
						{
							//restore old (default) displaymode
							GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().setDisplayMode(old);
						}
						//quit the game (dispose the window --> returns to the main Curved JFrame)
						window.dispose();
					}
				}
				//key released
				public void keyReleased (KeyEvent e)
				{
					if(running)
					{
						//movement keys --> set released flags
						for(int i=0;i<players;i++)
						{
							if (e.getKeyCode() == leftKeys[i])
							{
								left[i]=false;
							}
							else if (e.getKeyCode() == rightKeys[i])
							{
								right[i]=false;
							}
						}
					}
				}
			});

			//request focus to make sure key events are handled
			requestFocus();
	}

	//start a new game
	public void restart()
	{
		//reset the scores
		scores=new int[players];
		//reset the winner
		winner=0;
		//set flags
		theEnd=false;
		//start a new round
		nextRound();
	}

	//start new game round
	public void nextRound()
	{
		//reset the playfield
		playField=new boolean[704][594];
		//declare new snakes
		snakes = new Snake[players];
		//reset key events
		left = new boolean[players];
		right = new boolean[players];
		//initialize the snakes
		for(int i=0;i<players;i++)
		{
			snakes[i]=new Snake(colors[i],4,turnSpeed,holes);
		}
		//set running flag
		running=true;
		//set redraw flag
		redraw=true;
		//start new thread
		new Thread(this).start();
		//waitCounter --> wait this many game loops to start moving
		waitCounter=15;
		//reset alive
		alive=players;
	}

	//run method (Runnable interface) --> game loop
	public void run()
	{
		//continue the loop while the game is in running state
		while(running)
		{
			try
			{
				Thread.sleep(speed);
			}
			catch(InterruptedException e)
			{
				System.err.println("error: interrupted");
			}

			//hanle key events
			for(int i=0;i<players;i++)
			{
				if(left[i])
				{
					snakes[i].moveLeft();
				}
				else if(right[i])
				{
					snakes[i].moveRight();
				}
			}

			//move the snakes only is the waitcounter reached 0
			if(waitCounter<=0)
			{
				//update the snake's x and y cordinates
				for(int i=0; i<players; i++)
				{
					snakes[i].move();
				}
			}

			//repaint method --> calls paint
			repaint();
		}
	}

	//paint method
	public void paint(Graphics g)
	{
		//draw start screen - splash screen
		if(!started)
		{
			g.setColor(Color.orange);
			g.fillRect(0,0,800,600);
			g.setColor(bgColor);
			g.fillRect(3,3,794,594);
			g.setColor(Color.orange);
			g.setFont(new Font("monospaced",Font.BOLD,26));
			g.drawString("~ CurveD ~",280,210);
			g.setFont(new Font("monospaced",Font.BOLD,16));
			g.drawString("- press '" + KeyEvent.getKeyText(readyKey) + "' to continue -",220,240);
			for(int i=0;i<players;i++)
			{
				g.setColor(colors[i]);
				g.drawString("* snake "+(i+1) + ": '" + KeyEvent.getKeyText(leftKeys[i]) + "' = left, '" + KeyEvent.getKeyText(rightKeys[i])+"' = right" ,170,320+(i*30));
			}
			g.setColor(Color.white);
			g.setFont(new Font("monospaced",Font.PLAIN,14));
			g.drawString("a free opensource java game by www.knot.be",450,590);
		}
		//draw new playfield
		if(redraw)
		{
			g.setColor(Color.orange);
			g.fillRect(0,0,710,600);
			g.setColor(bgColor);
			g.fillRect(3,3,704,594);
			if(!started) //draw scoreboard
			{
				g.setFont(new Font("monospaced",Font.BOLD,14));
				g.setColor(Color.darkGray);
				g.fillRect(710,3,87,594);
				for (int j=0;j<players;j++)
				{
					g.setColor(colors[j]);
					g.drawString("***", 740,50+(j*60));
					g.setColor(Color.white);
					g.drawString("0 pts", 735,70+(j*60));
				}
				//set started flag
				started=true;
			}
			//unset redraw flag
			redraw=false;
		}
		//draw the Snakes
		if(snakes!=null)
		{
			if(waitCounter>0)
			{
				//make the startpoints of each snake blink
				if(waitCounter%2==0)
				{
					g.setColor(bgColor);
					g.fillRect(3,3,704,594);
				}
				else
				{
					for(int i=0; i<snakes.length; i++)
					{
						//get the color
						g.setColor(snakes[i].getColor());
						g.fillRect(snakes[i].getX(),snakes[i].getY(),snakes[i].getWidth(),snakes[i].getWidth());
					}
				}
				waitCounter--;
			}
			else
			{
				//draw each Snake in snakes[]
				for(int i=0; i<snakes.length; i++)
				{
					//get the color
					g.setColor(snakes[i].getColor());

					//only draw if necessary (if no snakehole)
					if(!snakes[i].isDead() && !snakes[i].dontDraw())
					{
						//draw the Snake
						g.fillRect(snakes[i].getX(),snakes[i].getY(),snakes[i].getWidth(),snakes[i].getWidth());

						//detect collision
						if(!addMove(snakes[i])) //this snake dies
						{
							//kill the snake
							snakes[i].kill();
							//decrease # of alive snakes
							alive--;
							//update the scores
							for (int j=0; j<players; j++)
							{
								//each living snake gets a point
								if(!snakes[j].isDead()) scores[j]++;
							}
							g.setFont(new Font("monospaced",Font.BOLD,14));
							//redraw the scoreboard
							g.setColor(Color.darkGray);
							g.fillRect(710,3,87,594);
							for (int j=0;j<players;j++)
							{
								g.setColor(colors[j]);
								g.drawString("***", 740,50+(j*60));
								g.setColor(Color.white);
								g.drawString(scores[j] + "pts", 735,70+(j*60));
								if(scores[j]>=maxScore) // j = winner
								{
									theEnd=true;
									running=false;
									winner=j;
								}
							}
							//if only 1 snake left
							if (alive<=1)
							{
								running=false;
								if(!theEnd)
								{
									g.setColor(bgColor);
									g.fillRect(206,224,320,24);
									g.setColor(Color.orange);
									g.setFont(new Font("monospaced",Font.BOLD,16));
									g.drawString("- press '" + KeyEvent.getKeyText(readyKey) + "' to continue -",220,240);
								}
							}
						}
					}
				}
			}
			//game over
			if(theEnd)
			{
				g.setColor(Color.orange);
				g.setFont(new Font("monospaced",Font.BOLD,26));
				g.drawString("~ Game Over ~",220,210);
				g.setFont(new Font("monospaced",Font.BOLD,16));
				g.drawString("'" + KeyEvent.getKeyText(readyKey) + "' = new game, '" + KeyEvent.getKeyText(quitKey) + "' = quit",180,280);
				g.setColor(colors[winner]);
				g.drawString("- player " + (winner+1) + " wins -",235,240);
			}
		}
	}

	//adds the next move of Snake s to the game
	//returns false when the Snake collides
	private boolean addMove(Snake s)
	{
		Point pnt;
		int x= s.getX() - 3;
		int y= s.getY() - 3;
		try
		{
			for(int i=x; i<=x+s.getWidth(); i++)
			{
				for(int j=y; j<=y+s.getWidth(); j++)
				{
					pnt=new Point(i,j);
					//if the position is occupied and it is not one of the latest moves
					//of this snake --> collision
					if(!s.contains(pnt) && playField[i][j]==true)
					{
						//--> collision (with a snake)
						return false;
					}
					//set occupied to true for this position
					playField[i][j]=true;
				}
			}
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			//--> collision (with a wall)
			return false;
		}
		return true;
	}

	//load the game properties
	private void loadProperties()
	{
		try
		{
			props=new Properties();

			//load the properties from the file 'properties'
			props.load(new FileInputStream("properties"));

			fullscreen=(new Boolean(props.getProperty("fullscreen"))).booleanValue();
			players=Integer.parseInt(props.getProperty("players"));
			maxScore=Integer.parseInt(props.getProperty("maxscore"));
			//colors and key configuration
			colors=new Color[players];
			leftKeys=new int[players];
			rightKeys=new int[players];
			for(int i=0;i<players;i++)
			{
				colors[i]=new Color(Integer.parseInt(props.getProperty("color"+(i+1))));
				leftKeys[i]=Integer.parseInt(props.getProperty("left"+(i+1)));
				rightKeys[i]=Integer.parseInt(props.getProperty("right"+(i+1)));
			}
			readyKey=Integer.parseInt(props.getProperty("ready"));
			quitKey=Integer.parseInt(props.getProperty("quit"));

			//gamespeed
			int i=Integer.parseInt(props.getProperty("speed"));
			switch(i)
			{
				case 0: speed=40; break;
				case 1: speed=30; break;
				default: speed=20;break;
			}
			turnSpeed=Integer.parseInt(props.getProperty("turn"));
			holes=Integer.parseInt(props.getProperty("hole"));
		}
		catch(Exception e)
		{
			System.err.println("\nAn error occured while loading 'properties'");
		};
	}
}